#ifndef XPP_HPP
#define XPP_HPP

#include "generic.hpp"

#include "atom.hpp"
#include "colormap.hpp"
#include "cursor.hpp"
#include "drawable.hpp"
#include "font.hpp"
#include "fontable.hpp"
#include "gcontext.hpp"
#include "pixmap.hpp"
#include "window.hpp"

#include "event.hpp"
#include "connection.hpp"

#endif // XPP_HPP
