#ifndef XPP_GENERIC_HPP
#define XPP_GENERIC_HPP

#include "generic/error.hpp"
#include "generic/event.hpp"
#include "generic/factory.hpp"
#include "generic/request.hpp"
#include "generic/resource.hpp"
#include "generic/extension.hpp"
#include "generic/signature.hpp"
#include "generic/reply_iterator.hpp"
#include "generic/iterator_traits.hpp"
#include "generic/input_iterator_adapter.hpp"

#endif // XPP_GENERIC_HPP
